const axios = require('axios');
const ethers = require('ethers');
const crypto = require('crypto');
const objectHash = require('object-hash');
const Promise = require('bluebird');

require('dotenv').config();

const verifySignature = require('./verify-signature');
const { getContractAddressByNetworkName } = require('./contract-config');
const { authorAddresses, authorPrivKeys, forum } = require('./db');
const DAOArtifact = require('../contractArtifacts/DAO.json');

// Each post allocates 30% of its reputation to citations
const PPM_TO_CITATIONS = 300000;

const network = process.env.ETH_NETWORK;
console.log('network:', network);
const getProvider = () => {
  switch (network) {
    case 'localhost':
      return ethers.getDefaultProvider('http://localhost:8545');
    case 'sepolia':
      return new ethers.InfuraProvider(
        network,
        process.env.INFURA_API_KEY,
      );
    default:
      throw new Error('Unknown network');
  }
};
const signer = new ethers.Wallet(process.env.ETH_PRIVATE_KEY, getProvider());
const getContract = (name) => new ethers.Contract(
  getContractAddressByNetworkName(process.env.ETH_NETWORK, name),
  DAOArtifact.abi,
  signer,
);
const dao = getContract('DAO');

const fetchWithRetry = async (url, retryDelay = 5000) => {
  let retry = false;
  let responseData;
  let requestError;
  const response = await axios.get(url, {
    headers: {
      'api-key': process.env.SEMANTIC_SCHOLAR_API_KEY,
    },
  }).catch(async (error) => {
    if (error.response?.status === 429) {
      // Rate limit
      retry = true;
      return;
    }
    // Some other error occurred
    requestError = error;
  });
  if (requestError) {
    throw requestError;
  } else if (retry) {
    console.log('retry delay (sec):', retryDelay / 1000);
    await new Promise((resolve) => {
      setTimeout(resolve, retryDelay);
    });
    responseData = await fetchWithRetry(url, retryDelay * 2);
  } else {
    responseData = response.data;
  }
  return responseData;
};

const fetchPaperInfo = async (paperId) => {
  const url = `https://api.semanticscholar.org/graph/v1/paper/${paperId}?`
  + 'fields=title,url,authors,references.title,references.url,references.authors';
  return fetchWithRetry(url);
};

const fetchAuthorPapers = async (authorId) => {
  const url = `https://api.semanticscholar.org/graph/v1/author/${authorId}/papers?`
  + 'fields=title,url,authors,references.title,references.url,references.authors';
  const responseData = await fetchWithRetry(url);
  // TODO: pagination
  return responseData.data;
};

const getOrCreateAuthor = async (paper) => Promise.mapSeries(
  paper.authors.filter((x) => !!x.authorId),
  async ({ authorId }) => {
  // Check if we already have an account for each author
    let authorAddress;
    let authorPrivKey;
    try {
      authorAddress = await authorAddresses.get(authorId);
    } catch (e) {
    // Probably not found
    }
    if (authorAddress) {
    // This should always succeed, so we don't use try/catch here
      authorPrivKey = await authorPrivKeys.get(authorAddress);
    } else {
    // Generate and store a new account
      const id = crypto.randomBytes(32).toString('hex');
      authorPrivKey = `0x${id}`;
      const wallet = new ethers.Wallet(authorPrivKey);
      authorAddress = wallet.address;
      await authorAddresses.put(authorId, authorAddress);
      await authorPrivKeys.put(authorAddress, authorPrivKey);
    }
    return {
      authorAddress,
      authorPrivKey,
    };
  },
);

const generatePost = async (paper) => {
  const authorsInfo = await getOrCreateAuthor(paper);
  if (!authorsInfo.length) {
    throw new Error('Paper has no authors with id');
  }
  const firstAuthorWallet = new ethers.Wallet(authorsInfo[0].authorPrivKey);
  const eachAuthorWeightPercent = Math.floor(1000000 / authorsInfo.length);
  const authors = authorsInfo.map(({ authorAddress }) => ({
    weightPPM: eachAuthorWeightPercent,
    authorAddress,
  }));
    // Make sure author weights sum to 100%
  const totalAuthorsWeight = authors.reduce((t, { weightPPM }) => t + weightPPM, 0);
  authors[0].weightPPM += 1000000 - totalAuthorsWeight;

  const content = `Semantic Scholar paper ${paper.paperId}
${paper.title}
HREF ${paper.url}`;

  const embeddedData = {
    semanticScholarPaperId: paper.paperId,
  };
  let contentToSign = content;
  if (embeddedData && Object.entries(embeddedData).length) {
    contentToSign += `\n\n${JSON.stringify(embeddedData, null, 2)}`;
  }
  const signature = firstAuthorWallet.signMessageSync(contentToSign);
  const verified = verifySignature({
    authors, content, signature, embeddedData,
  });
  if (!verified) {
    throw new Error('Signature verification failed');
  }
  const hash = objectHash({
    authors, content, signature, embeddedData,
  });
  return {
    hash, authors, content, signature, embeddedData,
  };
};

const importPaper = async (paper) => {
  console.log('references count:', paper.references.length);
  const { paperId } = paper;
  const references = paper.references.filter((x) => !!x.paperId);
  const eachCitationWeightPercent = Math.floor(PPM_TO_CITATIONS / references.length);
  const citations = await Promise.mapSeries(
    references,
    async (citedPaper) => {
      // We need to fetch this paper so we can generate the post we WOULD add to the forum.
      // That way, if we later add the cited paper to the blockchain it will have the correct hash.
      // The forum allows dangling citations to support this use case.
      const citedPost = await generatePost(citedPaper);
      return {
        weightPPM: eachCitationWeightPercent,
        targetPostId: citedPost.hash,
      };
    },
  );

  // Make sure citation weights sum to the designated total
  if (citations.length) {
    const totalCitationWeight = citations.reduce((t, { weightPPM }) => t + weightPPM, 0);
    citations[0].weightPPM += PPM_TO_CITATIONS - totalCitationWeight;
  }

  // Create a post for this paper
  const {
    hash, authors, content, signature, embeddedData,
  } = await generatePost(paper);

  // Write the new post to our database
  await forum.put(hash, {
    authors, content, signature, embeddedData, citations,
  });

  // Add the post to the form (on-chain)
  try {
    await dao.addPost(authors, hash, citations);
  } catch (e) {
    if (e.reason === 'A post with this contentId already exists') {
      console.log(`Post already added for paper ${paperId}`);
      return { paperId, postId: hash, alreadyAdded: true };
    }
    throw e;
  }

  console.log(`Added post to blockchain for paper ${paperId}`);
  return { paperId, postId: hash };
};

module.exports = async (req, res) => {
  const {
    body: {
      paperId,
      authorId,
    },
  } = req;

  if (paperId) {
    console.log(`importFromSS: paper ${paperId}`);
    const paper = await fetchPaperInfo(paperId);
    const result = await importPaper(paper);
    res.json(result);
  } else if (authorId) {
    console.log(`importFromSS: author ${authorId}`);
    const papers = await fetchAuthorPapers(authorId);
    console.log('papers count:', papers.length);
    const result = await Promise.mapSeries(papers, importPaper);
    console.log(`Added posts for ${result.length} papers by author ${authorId}`);
    res.json(result);
  }
};
