// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./Reputation.sol";
import "hardhat/console.sol";

struct Citation {
    int weightPPM;
    string targetPostId;
}

struct Author {
    uint weightPPM;
    address authorAddress;
}

struct Post {
    string id;
    address sender;
    Author[] authors;
    Citation[] citations;
    uint reputation;
    // TODO: timestamp
}

contract Forum is Reputation {
    mapping(string => Post) public posts;
    string[] public postIds;
    uint public postCount;
    mapping(string => mapping(string => int)) _edgeBalances;

    event PostAdded(string id);

    // Forum parameters
    // TODO: Make depth limit configurable; take as param in _onValidatePost callback
    uint depthLimit = 3;

    function addPost(
        Author[] calldata authors,
        string calldata contentId,
        Citation[] calldata citations
    ) external {
        require(authors.length > 0, "Post must include at least one author");
        postCount++;
        postIds.push(contentId);
        Post storage post = posts[contentId];
        require(
            post.authors.length == 0,
            "A post with this contentId already exists"
        );
        post.sender = msg.sender;
        post.id = contentId;
        uint authorTotalWeightPercent;
        for (uint i = 0; i < authors.length; i++) {
            authorTotalWeightPercent += authors[i].weightPPM;
            post.authors.push(authors[i]);
        }
        require(
            authorTotalWeightPercent == 1000000,
            "Author weights must sum to 1000000"
        );
        for (uint i = 0; i < citations.length; i++) {
            post.citations.push(citations[i]);
        }
        int totalCitationWeightPos;
        int totalCitationWeightNeg;
        for (uint i = 0; i < post.citations.length; i++) {
            int weight = post.citations[i].weightPPM;
            require(
                weight >= -1000000,
                "Each citation weight must be >= -1000000"
            );
            require(
                weight <= 1000000,
                "Each citation weight must be <= 1000000"
            );
            if (weight > 0) totalCitationWeightPos += weight;
            else totalCitationWeightNeg += weight;
        }
        require(
            totalCitationWeightPos <= 1000000,
            "Sum of positive citations must be <= 1000000"
        );
        require(
            totalCitationWeightNeg >= -1000000,
            "Sum of negative citations must be >= -1000000"
        );
        emit PostAdded(contentId);
    }

    function getPostAuthors(
        string calldata postId
    ) external view returns (Author[] memory) {
        Post storage post = posts[postId];
        return post.authors;
    }

    function _onValidatePost(string memory postId, uint amount) internal {
        _propagateReputation(postId, int(amount), false, 0);
    }

    function _handleCitation(
        string memory postId,
        Citation memory citation,
        int amount,
        bool initialNegative,
        uint depth
    ) internal returns (int outboundAmount) {
        outboundAmount = (amount * citation.weightPPM) / 1000000;
        if (bytes(citation.targetPostId).length == 0) {
            // Incineration
            require(
                outboundAmount >= 0,
                "Leaching from incinerator is forbidden"
            );
            _burn(address(this), uint(outboundAmount));
            return outboundAmount;
        }
        int balanceToOutbound = _edgeBalances[postId][citation.targetPostId];
        if (initialNegative) {
            if (outboundAmount < 0) {
                outboundAmount = outboundAmount > -balanceToOutbound
                    ? outboundAmount
                    : -balanceToOutbound;
            } else {
                outboundAmount = outboundAmount < -balanceToOutbound
                    ? outboundAmount
                    : -balanceToOutbound;
            }
        }
        int refund = _propagateReputation(
            citation.targetPostId,
            outboundAmount,
            initialNegative || (depth == 0 && citation.weightPPM < 0),
            depth + 1
        );
        outboundAmount -= refund;
        _edgeBalances[postId][citation.targetPostId] += outboundAmount;
    }

    function _distributeAmongAuthors(
        Post memory post,
        int amount
    ) internal returns (int refund) {
        int allocated;

        for (uint i = 0; i < post.authors.length; i++) {
            address authorAddress = post.authors[i].authorAddress;
            if (!isMember[authorAddress]) {
                members[memberCount++] = authorAddress;
                isMember[authorAddress] = true;
            }
        }
        for (uint i = 0; i < post.authors.length; i++) {
            Author memory author = post.authors[i];
            int share;
            if (i < post.authors.length - 1) {
                share = (amount * int(author.weightPPM)) / 1000000;
                allocated += share;
            } else {
                // For the last author, allocate the remainder.
                share = amount - allocated;
            }
            if (share > 0) {
                _update(address(this), author.authorAddress, uint(share));
                if (!isMember[author.authorAddress]) {
                    members[memberCount++] = author.authorAddress;
                    isMember[author.authorAddress] = true;
                }
            } else if (balanceOf(author.authorAddress) < uint(-share)) {
                // Author has already lost some REP gained from this post.
                // That means other DAO members have earned it for policing.
                // We need to refund the difference here to ensure accurate bookkeeping
                refund += share + int(balanceOf(author.authorAddress));
                _update(
                    author.authorAddress,
                    address(this),
                    balanceOf(author.authorAddress)
                );
            } else {
                _update(author.authorAddress, address(this), uint(-share));
            }
        }
    }

    function _propagateReputation(
        string memory postId,
        int amount,
        bool initialNegative,
        uint depth
    ) internal returns (int refundToInbound) {
        if (depth >= depthLimit) {
            return amount;
        }
        Post storage post = posts[postId];
        if (post.authors.length == 0) {
            // We most likely got here via a citation to a post that hasn't been added yet.
            // We support this scenario so that a citation graph can be imported one post at a time.
            return amount;
        }
        // Propagate negative citations first
        for (uint i = 0; i < post.citations.length; i++) {
            if (post.citations[i].weightPPM < 0) {
                amount -= _handleCitation(
                    postId,
                    post.citations[i],
                    amount,
                    initialNegative,
                    depth
                );
            }
        }
        // Now propagate positive citations
        for (uint i = 0; i < post.citations.length; i++) {
            if (post.citations[i].weightPPM > 0) {
                amount -= _handleCitation(
                    postId,
                    post.citations[i],
                    amount,
                    initialNegative,
                    depth
                );
            }
        }
        if (amount > 0) {
            _distributeAmongAuthors(post, amount);
            post.reputation += uint(amount);
        } else {
            if (int(post.reputation) + amount >= 0) {
                // Reduce the reputation of each author proportionately;
                // If any author has insufficient reputation, refund the difference.
                refundToInbound = _distributeAmongAuthors(post, amount);
                post.reputation -= uint(-amount);
            } else {
                // If we applied the full amount, the post's reputation would decrease below zero.
                refundToInbound = int(post.reputation) + amount;
                refundToInbound += _distributeAmongAuthors(
                    post,
                    -int(post.reputation)
                );
                post.reputation = 0;
            }
        }
    }
}
